"""
This module contains utility functions meant to be used in tests.
"""

import os

# Constants for export
CONTENT_TYPE = "Content-Type"
APPLICATION_JSON = "application/json"

LAMBDA_AUTH_EVENT_TYPE_REQUEST = "REQUEST"
LAMBDA_AUTH_EVENT_TYPE_TOKEN = "TOKEN"


def set_aws_environment_variables(
    aws_access_key_id: str,
    aws_secret_access_key: str,
    aws_security_token: str,
    aws_session_token: str,
    aws_default_region: str = "us-east-1",
    **kwargs,
):
    """
    Sets some environment variables used by AWS.
    """
    os.environ["AWS_ACCESS_KEY_ID"] = aws_access_key_id
    os.environ["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key
    os.environ["AWS_SECURITY_TOKEN"] = aws_security_token
    os.environ["AWS_SESSION_TOKEN"] = aws_session_token
    os.environ["AWS_DEFAULT_REGION"] = aws_default_region

    if kwargs:
        os.environ.update(kwargs)


def set_fake_aws_environment_variables(aws_default_region: str = "us-east-1", **kwargs):
    """
    Sets some environment variables used by AWS with fake values.
    This is meant to be used mainly for testing purposes, as some resources provided
    by the ``boto3`` library refuse to initialize when certain environment variables
    are empty.
    """
    set_aws_environment_variables(
        "testing", "testing", "testing", "testing", aws_default_region, **kwargs
    )


def get_fake_aws_lambda_context() -> dict:
    """
    Returns a dictionary containing fake values for information typically found in AWS
    Lambda Context objects. This is meant to be used mainly for testing purposes.
    """
    return {
        "aws_request_id": "testing",
        "log_group_name": "testing",
        "log_stream_name": "testing",
        "function_name": "testing",
        "function_version": "1",
        "invoked_function_arn": "testing",
        "memory_limit_in_mb": "128",
    }


def get_basic_aws_lambda_event(body=None, **kwargs) -> dict:
    """
    Returns a dictionary containing basic values for an AWS Lambda event.
    The keys present in the result dictionary are "body", whose value can be
    provided by the caller, and "headers", which is also a dictionary containing
    only the key "Content-Type" bound to value "application/json".
    """
    return {
        "body": body,
        "headers": {CONTENT_TYPE: APPLICATION_JSON},
        **kwargs,
    }


def get_aws_lambda_event_for_request_authorizer(
    method_arn="",
    resource="/",
    path="/",
    http_method="GET",
    headers: dict = None,
    multi_value_headers: dict = None,
    query_string_parameters: dict = None,
    multi_value_query_string_parameters: dict = None,
    path_parameters: dict = None,
    stage_variables: dict = None,
    **kwargs,
) -> dict:
    """
    Returns a dictionary with the typical structure of an event received by a
    request-based Lambda authorizer. This is meant to be used mainly for testing
    purposes.
    """
    if headers is None:
        headers = {}
    if multi_value_headers is None:
        multi_value_headers = {}
    if query_string_parameters is None:
        query_string_parameters = {}
    if multi_value_query_string_parameters is None:
        multi_value_query_string_parameters = {}
    if path_parameters is None:
        path_parameters = {}
    if stage_variables is None:
        stage_variables = {}

    return {
        "type": LAMBDA_AUTH_EVENT_TYPE_REQUEST,
        "methodArn": method_arn,
        "resource": resource,
        "path": path,
        "httpMethod": http_method,
        "headers": headers,
        "multiValueHeaders": multi_value_headers,
        "queryStringParameters": query_string_parameters,
        "multiValueQueryStringParameters": multi_value_query_string_parameters,
        "pathParameters": path_parameters,
        "stageVariables": stage_variables,
        **kwargs,
    }


def get_aws_lambda_event_for_token_authorizer(
    method_arn="",
    authorization_token="",
    **kwargs,
) -> dict:
    """
    Returns a dictionary with the typical structure of an event received by a
    token-based Lambda authorizer. This is meant to be used mainly for testing
    purposes.
    """
    return {
        "type": LAMBDA_AUTH_EVENT_TYPE_TOKEN,
        "methodArn": method_arn,
        "authorizationToken": authorization_token,
        **kwargs,
    }



"""Recreates the Python virtual environment for the project."""

import os
import shutil

# The directory where the Python virtual environment is stored
DEFAULT_PYTHON_VENV_DIR = "venv"
CUSTOM_PYTHON_VENV_DIR = DEFAULT_PYTHON_VENV_DIR
if os.getenv("AWS_COMMON_PYTHON_VENV_DIR"):
    CUSTOM_PYTHON_VENV_DIR = os.getenv("AWS_COMMON_PYTHON_VENV_DIR")
    print(f"Using Python virtual environment directory: {CUSTOM_PYTHON_VENV_DIR}")


def do_reset_venv(do_log_func, run_cmd_func, ignore_central_venv=False):
    """
    Recreate the Python virtual environment.
    """
    # If ignore_central_venv is True, do not use the centralized venv configuration
    if ignore_central_venv:
        do_log_func("Ignoring centralized virtual environment configuration.")
        venv_dir = DEFAULT_PYTHON_VENV_DIR
    else:
        venv_dir = CUSTOM_PYTHON_VENV_DIR

    do_log_func(f"*** Deleting all content under {venv_dir}...")
    shutil.rmtree(venv_dir, ignore_errors=True)

    do_log_func("*** Recreating Python virtual environment...")
    run_cmd_func(["python3.11", "-m", "venv", venv_dir])

    activate_script = os.path.join(venv_dir, "bin", "activate")

    # warn the user to activate the virtual environment
    print("*** Virtual environment recreated! Please activate it by running:")
    print(f"source {activate_script}")



"""
This module contains the exceptions that should be handled internally
and not exposed to the user.
"""


class NonUserFacingException(Exception):
    """
    Exception raised for errors that should be handled internally
    and not exposed to the user.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message="An internal error occurred.",
    ):
        self.message = message
        super().__init__(self.message)



"""
This module is used to install all the project requirements.
"""

import os
import shutil


def _purge_pip_cache(do_log_func, run_cmd_func):
    """
    Purge the pip cache to avoid potential installation issues.
    """
    run_cmd_func(["python3.11", "-m", "pip", "cache", "purge"])
    do_log_func("*** Purged pip cache.")


def _remove_pip_selfcheck(do_log_func):
    """
    Remove the pip selfcheck directory.
    """
    selfcheck_dir = os.path.expanduser("~/.cache/pip/selfcheck/")
    if os.path.exists(selfcheck_dir):
        shutil.rmtree(selfcheck_dir)
    do_log_func("*** Removed pip cache selfcheck directory.")


def _upgrade_pip(do_log_func, run_cmd_func):
    """
    Upgrade pip to the latest version.
    """
    run_cmd_func(["pip", "install", "--upgrade", "pip", "--quiet"])
    do_log_func("*** Upgraded pip.")


def _install_dev_requirements(do_log_func, run_cmd_func, target=None):
    """
    Installs or upgrades essential module dependencies directly from
    the aws-common project.

    Parameters:
    do_log_func (function): A logging function to record progress messages.
    run_cmd_func (function): A function to execute shell commands.
    target (str, optional): The target directory where the dependencies
                            should be installed. Defaults to None.
    """
    # Install dependencies from the requirements-dev.txt file located
    # in the aws-common project.
    # The libraries specified in the requirements.dev.txt will be installed only
    # in the local Python virtual environments to ensure the scripts are executed.
    # The requirements-dev.txt file is expected to be in the
    # parent directory of this script.
    aws_common_req_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "requirements-dev.txt"
    )
    do_log_func(f"*** Installing/upgrading from {aws_common_req_path}...")
    cmd_list = ["pip", "install", "-r", aws_common_req_path, "--quiet"]
    if target:
        cmd_list += ["--target", target]
    run_cmd_func(cmd_list)


def _install_from_aws_common(do_log_func, run_cmd_func, target=None):
    """
    Installs or upgrades essential module dependencies directly from
    the aws-common project.

    Parameters:
    do_log_func (function): A logging function to record progress messages.
    run_cmd_func (function): A function to execute shell commands.
    target (str, optional): The target directory where the dependencies
                            should be installed. Defaults to None.
    """
    # Install dependencies from the requirements.txt file located
    # in the aws-common project.
    # The requirements.txt file is expected to be in the
    # parent directory of this script.
    aws_common_req_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "requirements.txt"
    )
    do_log_func(f"*** Installing/upgrading from {aws_common_req_path}...")

    cmd_list = ["pip", "install", "-r", aws_common_req_path, "--quiet"]
    if target:
        cmd_list += ["--target", target]
    run_cmd_func(cmd_list)

    # Install the aws-common project directly from the Git repository.
    # This is necessary while the aws-common project is not published on PyPI.
    # TODO: #26 Include aws-common project in the PyPI repository.
    git_repo_url = "https://github.com/LARS-Org/aws-common"
    do_log_func(f"*** Installing/upgrading {git_repo_url} (will be quiet)...")

    cmd_list = ["pip", "install", "--upgrade", f"git+{git_repo_url}", "--quiet"]
    if target:
        cmd_list += ["--target", target]
    run_cmd_func(cmd_list)

    # Log completion of the installation/upgrade process.
    do_log_func("*** Installation/upgrade completed successfully.")


def _install_requirements_recursively(do_log_func, run_cmd_func):
    """
    Recursively traverse the project directory and install all requirements.txt files.
    """
    pip_requirements_file_list = ["requirements.txt", "requirements-dev.txt"]

    for root, dirs, files in os.walk("."):
        # Skip certain directories
        if any(
            skip in root
            for skip in [
                ".aws-sam",
                "venv",
                ".git",
                ".pytest",
                "lib/python",
                "tests/lib/python",
                "experiments",
                "__pycache__",
                "node_modules",
                "cdk.out",
                "packages",
                "lib/python3.11/site-packages",
                "venv",
            ]
        ):
            continue

        do_log_func(f"*** Processing {root}...")

        # exclude the __init__.py files
        files = set(files) - {"__init__.py"}

        for pip_requirements_file in pip_requirements_file_list:
            if pip_requirements_file in files:
                pip_requirements_path = os.path.join(root, pip_requirements_file)
                do_log_func(f"Installing {pip_requirements_path} (will be quiet)...")
                run_cmd_func(["pip", "install", "-r", pip_requirements_path, "--quiet"])

        # Special handling for AWS Lambda Functions
        if files and root.startswith(("lambda", "./lambda", "tests", "./tests")):
            packages_dir = os.path.join(root, "packages")
            # remove the packages directory if it exists
            if os.path.exists(packages_dir):
                shutil.rmtree(packages_dir)
            # create the packages directory
            os.makedirs(packages_dir)

            # install the git repositories in the packages directory
            _install_from_aws_common(do_log_func, run_cmd_func, target=packages_dir)

            for pip_requirements_file in pip_requirements_file_list:
                if pip_requirements_file in files:
                    do_log_func(
                        f"Installing {pip_requirements_path} (will be quiet)..."
                    )
                    # install the requirements in the packages directory
                    run_cmd_func(
                        [
                            "pip",
                            "install",
                            "-r",
                            pip_requirements_path,
                            "--target",
                            packages_dir,
                            "--upgrade",
                            "--quiet",
                        ]
                    )

        # Special handling for AWS Lambda Layers
        if root.startswith(("layers", "./layers")):
            run_cmd_func(
                [
                    "pip",
                    "install",
                    "-r",
                    pip_requirements_path,
                    "--target",
                    os.path.join(root, "lib/python3.11/site-packages"),
                    "--upgrade",
                    "--quiet",
                ]
            )


def _install_other_packages(do_log_func, run_cmd_func):
    """
    Install other pip packages required by the project.
    """
    other_packages = ["pylint", "black", "isort"]

    for package in other_packages:
        do_log_func(f"*** Installing/upgrading {package} (will be quiet)...")
        run_cmd_func(["pip", "install", "--upgrade", package, "--quiet"])


def do_install_req(do_log_func, run_cmd_func):
    """
    Main function to install all project requirements.
    """
    _purge_pip_cache(do_log_func, run_cmd_func)
    _remove_pip_selfcheck(do_log_func)
    _upgrade_pip(do_log_func, run_cmd_func)
    _install_dev_requirements(do_log_func, run_cmd_func)
    _install_from_aws_common(do_log_func, run_cmd_func)
    _install_requirements_recursively(do_log_func, run_cmd_func)
    _install_other_packages(do_log_func, run_cmd_func)
    do_log_func("*** All done!!!")



"""
A module that contains utility functions to be used in API Gateway-related tasks.
"""

# Constants for export
ACTION_EXECUTE_API_INVOKE = "execute-api:Invoke"
EFFECT_ALLOW = "Allow"
EFFECT_DENY = "Deny"

PRINCIPAL_ID = "principalId"
POLICY_DOCUMENT = "policyDocument"
ACTION = "Action"
EFFECT = "Effect"
RESOURCE = "Resource"


def generate_auth_response_as_iam_policy(
    principal_id,
    action=ACTION_EXECUTE_API_INVOKE,
    effect="",
    resource="",
    response_context: dict | None = None,
) -> dict:
    """
    Given the input parameters describing the response to an authorization request,
    formats the supplied information as an IAM policy dictionary and returns it.
    Designed to be used in Lambda functions underlying API Gateway Lambda authorizers.
    See:
      https://docs.aws.amazon.com/apigateway/latest/developerguide/apigateway-use-lambda-authorizer.html#api-gateway-lambda-authorizer-lambda-function-create

    Parameters
    ----------
    - principal_id: the ID of the principal (user, role, etc.) that the authorization
      request refers to;
    - action: the action that the authorization request refers to (usually
      "execute-api:Invoke");
    - effect: the effect of the IAM policy (either "Allow" or "Deny");
    - resource: the resource that the authorization request refers to. This is usually
      the ARN of a method execution descriptor over a resource of an API (e.g.,
      "arn:aws:execute-api:us-east-1:123456789012:ivdtdhp7b5/ESTestInvoke-stage/GET/");
    - response_context: a dictionary containing additional information to be sent back
      to the initiator of the authorization request.
    """

    auth_response = {"principalId": principal_id}

    if effect and resource:
        auth_response["policyDocument"] = {
            "Version": "2012-10-17",
            "Statement": [{"Action": action, "Effect": effect, "Resource": resource}],
        }

    if response_context:
        auth_response["context"] = response_context

    return auth_response


def generate_allow_auth_response(
    principal_id,
    action=ACTION_EXECUTE_API_INVOKE,
    resource="",
    response_context: dict | None = None,
):
    """
    A shortcut method that generates an IAM policy allowing the principal (user, role,
    etc.) with the given ID to proceed with an action on a resource.
    Designed to be used in Lambda functions underlying API Gateway Lambda authorizers.
    """

    return generate_auth_response_as_iam_policy(
        principal_id, action, EFFECT_ALLOW, resource, response_context
    )


def generate_deny_auth_response(
    principal_id,
    action=ACTION_EXECUTE_API_INVOKE,
    resource="",
    response_context: dict | None = None,
):
    """
    A shortcut method that generates an IAM policy denying the principal (user, role,
    etc.) with the given ID to proceed with an action on a resource.
    Designed to be used in Lambda functions underlying API Gateway Lambda authorizers.
    """

    return generate_auth_response_as_iam_policy(
        principal_id, action, EFFECT_DENY, resource, response_context
    )
